from pye57 import libe57
from pye57.libe57 import NodeType

import numpy as np

def get_fields(node):
    return [node.get(id_).elementName() for id_ in range(node.childCount())]


def get_node(node, name):
    cast = {
        NodeType.E57_BLOB: libe57.BlobNode,
        NodeType.E57_COMPRESSED_VECTOR: libe57.CompressedVectorNode,
        NodeType.E57_FLOAT: libe57.FloatNode,
        NodeType.E57_INTEGER: libe57.IntegerNode,
        NodeType.E57_SCALED_INTEGER: libe57.ScaledIntegerNode,
        NodeType.E57_STRING: libe57.StringNode,
        NodeType.E57_STRUCTURE: libe57.StructureNode,
        NodeType.E57_VECTOR: libe57.VectorNode
    }
    n = node.get(name)
    return cast[n.type()](n)

def convert_spherical_to_cartesian(rae):
    """
    Converts spherical(rae) to cartesian(xyz), where rae = range, azimuth(theta), 
    elevation(phi). Where range is in meters and angles are in radians.
    
    Reference for formula: http://www.libe57.org/bestCoordinates.html (Note: the 
    formula is different from the one online, so please use formula at the above reference)
    """
    range_ = rae[:, :1]
    theta = rae[:, 1:2]
    phi =  rae[:, 2:3]
    range_cos_phi = range_ * np.cos(phi)
    return np.concatenate((
        range_cos_phi * np.cos(theta),
        range_cos_phi * np.sin(theta),
        range_ * np.sin(phi)
    ), axis=1)


def copy_node(node, dest_image):
    compressed_node_pairs = []
    blob_node_pairs = []

    out_node = None
    # 'Element' Types
    if (isinstance(node, libe57.FloatNode)):
        out_node = libe57.FloatNode(
            dest_image,
            value=node.value(),
            precision=node.precision(),
            minimum=node.minimum(),
            maximum=node.maximum())

    elif (isinstance(node, libe57.IntegerNode)):
        out_node = libe57.IntegerNode(
            dest_image,
            value=node.value(),
            minimum=node.minimum(),
            maximum=node.maximum())

    elif (isinstance(node, libe57.ScaledIntegerNode)):
        out_node = libe57.ScaledIntegerNode(
            dest_image,
            node.rawValue(),
            minimum=node.minimum(),
            maximum=node.maximum(),
            scale=node.scale(),
            offset=node.offset())

    elif (isinstance(node, libe57.StringNode)):
        out_node = libe57.StringNode(
            dest_image,
            node.value())

    elif (isinstance(node, libe57.BlobNode)):
        out_node = libe57.BlobNode(dest_image, node.byteCount())
        blob_node_pairs.append({ 'in': node, 'out': out_node })

    # 'Container' Types
    elif (isinstance(node, libe57.CompressedVectorNode)):
        in_prototype = libe57.StructureNode(node.prototype())
        out_prototype, _, _  = copy_node(in_prototype, dest_image)
        out_codecs, _, _ = copy_node(node.codecs(), dest_image)

        out_node = libe57.CompressedVectorNode(dest_image, out_prototype, out_codecs)

        compressed_node_pairs.append({
            'in': node,
            'out': out_node
        })

    elif isinstance(node, libe57.StructureNode):
        out_node = libe57.StructureNode(dest_image)
        for i in range(node.childCount()):
            in_child = get_node(node, i)
            in_child_name = in_child.elementName()
            out_child, out_child_compressed_node_pairs, out_child_blob_node_pairs = copy_node(in_child, dest_image)

            out_node.set(in_child_name, out_child)
            compressed_node_pairs.extend(out_child_compressed_node_pairs)
            blob_node_pairs.extend(out_child_blob_node_pairs)

    elif isinstance(node, libe57.VectorNode):
        out_node = libe57.VectorNode(dest_image, allowHeteroChildren=node.allowHeteroChildren())
        for i in range(node.childCount()):
            in_child = get_node(node, i)
            in_child_name = f'{i}'
            out_child, out_child_compressed_node_pairs, out_child_blob_node_pairs = copy_node(in_child, dest_image)

            out_node.append(out_child)
            compressed_node_pairs.extend(out_child_compressed_node_pairs)
            blob_node_pairs.extend(out_child_blob_node_pairs)

    return out_node, compressed_node_pairs, blob_node_pairs


import pytest
import os
import time

import numpy as np

import pye57
from pye57 import libe57
from pye57.utils import get_node, copy_node

try:
    from exceptions import WindowsError
except ImportError:
    class WindowsError(OSError):
        pass


def test_hi():
    assert libe57.__doc__


def sample_data(*args):
    here = os.path.split(__file__)[0]
    return os.path.join(here, "test_data", *args)


def delete_retry(path):
    try:
        if os.path.exists(path):
            os.remove(path)
    except WindowsError:
        time.sleep(0.1)
        if os.path.exists(path):
            os.remove(path)


@pytest.fixture
def e57_path():
    return sample_data("test.e57")


@pytest.fixture
def e57_spherical_path():
    return sample_data("testSpherical.e57")


@pytest.fixture
def e57_with_data_and_images_path():
    # From http://www.libe57.org/data.html
    return sample_data("pumpAVisualReferenceImage.e57")


@pytest.fixture
def temp_e57_write(request):
    path = sample_data("test_write.e57")
    request.addfinalizer(lambda: delete_retry(path))
    return path


@pytest.fixture
def image_and_points(e57_path):
    f = libe57.ImageFile(e57_path, mode="r")
    scan_0 = libe57.StructureNode(libe57.VectorNode(f.root().get("/data3D")).get(0))
    points = libe57.CompressedVectorNode(scan_0.get("points"))
    return f, points


def test_constants():
    assert libe57.CHECKSUM_POLICY_NONE == 0
    assert libe57.CHECKSUM_POLICY_SPARSE == 25
    assert libe57.CHECKSUM_POLICY_HALF == 50
    assert libe57.CHECKSUM_POLICY_ALL == 100
    assert libe57.E57_INT8_MIN == -128
    assert libe57.E57_INT8_MAX == 127
    assert libe57.E57_INT16_MIN == -32768
    assert libe57.E57_INT16_MAX == 32767
    assert libe57.E57_INT32_MIN == -2147483647 - 1
    assert libe57.E57_INT32_MAX == 2147483647
    assert libe57.E57_INT64_MIN == -9223372036854775807 - 1
    assert libe57.E57_INT64_MAX == 9223372036854775807
    assert libe57.E57_UINT8_MIN == 0
    assert libe57.E57_UINT8_MAX == 255
    assert libe57.E57_UINT16_MIN == 0
    assert libe57.E57_UINT16_MAX == 65535
    assert libe57.E57_UINT32_MIN == 0
    assert libe57.E57_UINT32_MAX == 4294967295
    assert libe57.E57_UINT64_MIN == 0
    assert libe57.E57_UINT64_MAX == 18446744073709551615


def test_open_imagefile(e57_path):
    f = libe57.ImageFile(e57_path, mode="r")
    assert f.isOpen()
    f.close()


def test_open_imagefile_write(temp_e57_write):
    f = libe57.ImageFile(temp_e57_write, mode="w")
    assert f.isOpen()
    f.close()


def test_e57_mode_error(temp_e57_write):
    with pytest.raises(ValueError):
        f = pye57.E57(temp_e57_write, mode="pasta")


def test_get_structure_names(e57_path):
    f = libe57.ImageFile(e57_path, "r")
    root = f.root()
    names = []
    for id_ in range(root.childCount()):
        names.append(root.get(id_).pathName())
    assert names == ['/formatName', '/guid', '/versionMajor', '/versionMinor', '/e57LibraryVersion',
                     '/coordinateMetadata', '/creationDateTime', '/data3D', '/images2D']


def test_get_data3d_nodes(e57_path):
    f = libe57.ImageFile(e57_path, "r")
    root = f.root()
    node = root.get("data3D")
    data3d = libe57.VectorNode(node)
    scan_count = data3d.childCount()
    assert scan_count == 4
    for scan_id in range(scan_count):
        assert isinstance(data3d.get(scan_id), libe57.Node)


def test_get_read_data3d(e57_path):
    f = libe57.ImageFile(e57_path, "r")
    scan_0 = libe57.StructureNode(libe57.VectorNode(f.root().get("/data3D")).get(0))
    points = libe57.CompressedVectorNode(scan_0.get("points"))
    assert points.childCount() == 281300


def test_source_dest_buffers(e57_path):
    f = libe57.ImageFile(e57_path, "r")
    capacity = 1000
    types = list("bBhHlLq?fd")
    sizes = [1, 1, 2, 2, 4, 4, 8, 1, 4, 8]
    buffers = libe57.VectorSourceDestBuffer()
    for t in types:
        data = np.zeros(capacity, t)
        sdb = libe57.SourceDestBuffer(f, "something", data, capacity, True, True)
        buffers.append(sdb)

    for t, sdb, size, in zip(types, buffers, sizes):
        assert sdb.pathName() == "something"
        assert sdb.capacity() == capacity
        assert sdb.stride() == size
        assert sdb.doScaling()
        assert sdb.doConversion()


def test_unsupported_point_field(temp_e57_write):
    with pye57.E57(temp_e57_write, mode="w") as f:
        with pytest.raises(ValueError):
            data = {"cartesianX": np.random.rand(10),
                    "bananas": np.random.rand(10)}
            f.write_scan_raw(data)


def test_source_dest_buffers_raises(e57_path):
    f = libe57.ImageFile(e57_path, "r")
    capacity = 1000
    data = np.zeros(capacity, "i")
    with pytest.raises(ValueError):
        libe57.SourceDestBuffer(f, "something", data, capacity, True, True)


def test_read_points_x(image_and_points):
    imf, points = image_and_points
    bufs = libe57.VectorSourceDestBuffer()
    capacity = 10000
    X = np.zeros(capacity, "f")
    bufs.append(libe57.SourceDestBuffer(imf, "cartesianX", X, capacity, True, True))
    data_reader = points.reader(bufs)
    size = data_reader.read()
    assert size == capacity
    assert not np.all(np.zeros(capacity, "f") == X)


def test_index_out_of_range(e57_path):
    f = libe57.ImageFile(e57_path, "r")
    with pytest.raises(IndexError):
        scan = f.root()["data3D"][-1]
    with pytest.raises(IndexError):
        scan = f.root()["data3D"][5]
    scan_0 = f.root()["data3D"][0]
    with pytest.raises(IndexError):
        r = scan_0["pose"]["rotation"][-1]
    with pytest.raises(IndexError):
        r = scan_0["pose"]["rotation"][5]


def test_read_header(e57_path):
    f = libe57.ImageFile(e57_path, "r")
    data3d = f.root()["data3D"]
    headers = pye57.ScanHeader.from_data3d(data3d)
    fields = ['cartesianX', 'cartesianY', 'cartesianZ', 'intensity', 'rowIndex', 'columnIndex', 'cartesianInvalidState']
    for header in headers:
        assert fields == header.point_fields
    assert headers[0].pretty_print()
    scan_0_rot = [[-0.4443, 0.8958, 0.],
                  [-0.8958, -0.4443, 0.],
                  [0., 0., 1.]]
    assert np.allclose(scan_0_rot, headers[0].rotation_matrix, atol=1e-3)
    scan_0_tra = [301336.23199, 5042597.23676, 15.46649]
    assert np.allclose(scan_0_tra, headers[0].translation)


def test_read_xyz(e57_path):
    e57 = pye57.E57(e57_path)
    xyz = e57.read_scan(0)
    assert np.any(xyz)


def test_read_header_spherical(e57_spherical_path):
    f = libe57.ImageFile(e57_spherical_path, "r")
    data3d = f.root()["data3D"]
    headers = pye57.ScanHeader.from_data3d(data3d)
    fields = ['sphericalRange', 'sphericalAzimuth', 'sphericalElevation', 'intensity', 'colorRed', 'colorGreen', 'colorBlue', 'sphericalInvalidState']
    for header in headers:
        assert fields == header.point_fields
    assert headers[0].pretty_print()


def test_read_xyz_spherical(e57_spherical_path):
    e57 = pye57.E57(e57_spherical_path)
    xyz = e57.read_scan(0)
    assert np.any(xyz)
    

def test_read_raw(e57_path):
    e57 = pye57.E57(e57_path)
    header = e57.get_header(0)
    fields = header.point_fields
    data = e57.read_scan_raw(0)
    assert sorted(fields) == sorted(data.keys())
    assert np.any(data["cartesianX"])
    assert len(data["cartesianX"]) == header.point_count


def test_read_write_single_scan(e57_path, temp_e57_write):
    e57 = pye57.E57(e57_path)
    header_source = e57.get_header(0)
    with pye57.E57(temp_e57_write, mode="w") as e57_write:
        raw_data_0 = e57.read_scan_raw(0)
        e57_write.write_scan_raw(raw_data_0, rotation=header_source.rotation, translation=header_source.translation)
    scan_0 = pye57.E57(e57_path).read_scan_raw(0)
    written = pye57.E57(temp_e57_write)
    header = written.get_header(0)
    assert np.allclose(header.rotation, header_source.rotation)
    assert np.allclose(header.translation, header_source.translation)
    scan_0_written = written.read_scan_raw(0)
    fields = "cartesianX cartesianY cartesianZ intensity rowIndex columnIndex cartesianInvalidState".split()
    for field in fields:
        assert np.allclose(scan_0[field], scan_0_written[field])

    scan_0 = e57.read_scan(0)
    scan_0_written = written.read_scan(0)
    for field in scan_0:
        assert np.allclose(scan_0[field], scan_0_written[field])


def test_copy_file(e57_path, temp_e57_write):
    e57 = pye57.E57(e57_path)
    with pye57.E57(temp_e57_write, mode="w") as f:
        for scan_id in range(e57.scan_count):
            header = e57.get_header(scan_id)
            data = e57.read_scan_raw(scan_id)
            f.write_scan_raw(data, scan_header=header)
            header_written = f.get_header(scan_id)
            assert header_written.guid
            assert header_written.temperature == header_written.temperature
            assert header_written.relativeHumidity == header_written.relativeHumidity
            assert header_written.atmosphericPressure == header_written.atmosphericPressure
            assert header_written.rowMinimum == header.rowMinimum
            assert header_written.rowMaximum == header.rowMaximum
            assert header_written.columnMinimum == header.columnMinimum
            assert header_written.columnMaximum == header.columnMaximum
            assert header_written.returnMinimum == header.returnMinimum
            assert header_written.returnMaximum == header.returnMaximum
            assert header_written.intensityMinimum == header.intensityMinimum
            assert header_written.intensityMaximum == header.intensityMaximum
            assert header_written.xMinimum == header.xMinimum
            assert header_written.xMaximum == header.xMaximum
            assert header_written.yMinimum == header.yMinimum
            assert header_written.yMaximum == header.yMaximum
            assert header_written.zMinimum == header.zMinimum
            assert header_written.zMaximum == header.zMaximum
            assert np.allclose(header_written.rotation, header.rotation)
            assert np.allclose(header_written.translation, header.translation)
            assert header_written.acquisitionStart_dateTimeValue == header.acquisitionStart_dateTimeValue
            assert header_written.acquisitionStart_isAtomicClockReferenced == header.acquisitionStart_isAtomicClockReferenced
            assert header_written.acquisitionEnd_dateTimeValue == header.acquisitionEnd_dateTimeValue
            assert header_written.acquisitionEnd_isAtomicClockReferenced == header.acquisitionEnd_isAtomicClockReferenced
            # todo: point groups
            # header.pointGroupingSchemes["groupingByLine"]["idElementName"].value()
            # header.pointGroupingSchemes["groupingByLine"]["groups"]

        assert f.scan_count == e57.scan_count


def test_read_color_absent(e57_path):
    e57 = pye57.E57(e57_path)
    with pytest.raises(ValueError):
        data = e57.read_scan(0, colors=True)


def test_scan_position(e57_path):
    e57 = pye57.E57(e57_path)
    assert np.allclose(e57.scan_position(3), np.array([[3.01323456e+05, 5.04260184e+06, 1.56040279e+01]]))


BUFFER_TYPES = {
    libe57.FloatNode: 'd',
    libe57.IntegerNode: 'l',
    libe57.ScaledIntegerNode: 'd'
}


def make_buffer(node: libe57.Node, capacity: int):
    node_type = type(node)
    if node_type not in BUFFER_TYPES:
        raise ValueError("Unsupported field type!")

    buffer_type = BUFFER_TYPES[node_type]

    np_array = np.empty(capacity, buffer_type)
    buffer = libe57.SourceDestBuffer(
        node.destImageFile(), node.elementName(), np_array, capacity, True, True)
    return np_array, buffer


def make_buffers(node: libe57.StructureNode, capacity: int):
    data = {}
    buffers = libe57.VectorSourceDestBuffer()
    for i in range(node.childCount()):
        field = get_node(node, i)
        d, b = make_buffer(field, capacity)
        data[field.elementName()] = d
        buffers.append(b)
    return data, buffers


def copy_compressed_vector_data(in_node: libe57.Node, out_node: libe57.Node):
    chunk_size = 100000

    in_prototype = libe57.StructureNode(in_node.prototype())
    out_prototype = libe57.StructureNode(out_node.prototype())

    in_data, in_buffers = make_buffers(in_prototype, chunk_size)
    out_data, out_buffers = make_buffers(out_prototype, chunk_size)

    in_reader = in_node.reader(in_buffers)
    out_writer = out_node.writer(out_buffers)

    n_points = in_node.childCount()
    current_index = 0
    while current_index != n_points:
        current_chunk = min(n_points - current_index, chunk_size)

        in_reader.read()
        for field in in_data:
            out_data[field][:current_chunk] = in_data[field][:current_chunk]

        out_writer.write(current_chunk)

        current_index += current_chunk

    in_reader.close()
    out_writer.close()


def copy_blob_data(in_node, out_node):
    chunk_size = 100000

    byte_count = in_node.byteCount()
    blob_buffer = np.empty(chunk_size, np.ubyte)
    current_index = 0
    while current_index != byte_count:
        current_chunk = min(byte_count - current_index, chunk_size)

        in_node.read(blob_buffer, current_index, current_chunk)
        out_node.write(blob_buffer, current_index, current_chunk)

        current_index += current_chunk


def test_clone_e57(e57_with_data_and_images_path, temp_e57_write):

    in_image = libe57.ImageFile(e57_with_data_and_images_path, "r")
    out_image = libe57.ImageFile(temp_e57_write, "w")

    for i in range(in_image.extensionsCount()):
        out_image.extensionsAdd(
            in_image.extensionsPrefix(i),
            in_image.extensionsUri(i)
        )
    
    in_root = in_image.root()
    out_root = out_image.root()

    compressed_node_pairs = []
    for i in range(in_root.childCount()):
        in_child = get_node(in_root, i)
        in_child_name = in_child.elementName()
        out_child, out_child_compressed_node_pairs, out_child_blob_node_pairs = copy_node(in_child, out_image)

        out_root.set(in_child_name, out_child)
        compressed_node_pairs.extend(out_child_compressed_node_pairs)
    
    for compressed_node_pair in compressed_node_pairs:
        copy_compressed_vector_data(compressed_node_pair['in'], compressed_node_pair['out'])

    for blob_node_pair in out_child_blob_node_pairs:
        copy_blob_data(blob_node_pair['in'], blob_node_pair['out'])

    in_image.close()
    out_image.close()



import numpy as np
from pyquaternion import Quaternion

from pye57 import libe57
from pye57.utils import get_fields, get_node

class ScanHeader:
    def __init__(self, scan_node):
        self.node = scan_node
        points = self.node["points"]
        self.point_fields = get_fields(libe57.StructureNode(points.prototype()))
        self.scan_fields = get_fields(self.node)

    @classmethod
    def from_data3d(cls, data3d):
        return [cls(scan) for scan in data3d]

    def has_pose(self):
        return self.node.isDefined("pose")

    @property
    def point_count(self):
        return self.points.childCount()

    @property
    def rotation_matrix(self) -> np.array:
        q = Quaternion([e.value() for e in self.node["pose"]["rotation"]])
        return q.rotation_matrix

    @property
    def rotation(self) -> np.array:
        q = Quaternion([e.value() for e in self.node["pose"]["rotation"]])
        return q.elements

    @property
    def translation(self):
        return np.array([e.value() for e in self.node["pose"]["translation"]])

    def pretty_print(self, node=None, indent=""):
        if node is None:
            node = self.node
        lines = []
        for field in get_fields(node):
            child_node = node[field]
            value = ""
            if hasattr(child_node, "value"):
                value = ": %s" % child_node.value()
            lines.append(indent + str(child_node) + value)
            if isinstance(child_node, libe57.StructureNode):
                lines += self.pretty_print(child_node, indent + "    ")
        return lines

    def __getitem__(self, item):
        return self.node[item]
    
    def get_coordinate_system(self, COORDINATE_SYSTEMS):
        if all(x in self.point_fields for x in COORDINATE_SYSTEMS.CARTESIAN.value):
            coordinate_system = COORDINATE_SYSTEMS.CARTESIAN
        elif all(x in self.point_fields for x in COORDINATE_SYSTEMS.SPHERICAL.value):
            coordinate_system = COORDINATE_SYSTEMS.SPHERICAL
        else:
            raise Exception(f"Scans coordinate system not supported, unsupported point field {self.point_fields}")
        return coordinate_system            

    @property
    def guid(self):
        return self["guid"].value()

    @property
    def temperature(self):
        return self["temperature"].value()

    @property
    def relativeHumidity(self):
        return self["relativeHumidity"].value()

    @property
    def atmosphericPressure(self):
        return self["atmosphericPressure"].value()

    @property
    def indexBounds(self):
        return self["indexBounds"]

    @property
    def rowMinimum(self):
        return self.indexBounds["rowMinimum"].value()

    @property
    def rowMaximum(self):
        return self.indexBounds["rowMaximum"].value()

    @property
    def columnMinimum(self):
        return self.indexBounds["columnMinimum"].value()

    @property
    def columnMaximum(self):
        return self.indexBounds["columnMaximum"].value()

    @property
    def returnMinimum(self):
        return self.indexBounds["returnMinimum"].value()

    @property
    def returnMaximum(self):
        return self.indexBounds["returnMaximum"].value()

    @property
    def intensityLimits(self):
        return self["intensityLimits"]

    @property
    def intensityMinimum(self):
        return self.intensityLimits["intensityMinimum"].value()

    @property
    def intensityMaximum(self):
        return self.intensityLimits["intensityMaximum"].value()

    @property
    def cartesianBounds(self):
        return self["cartesianBounds"]

    @property
    def xMinimum(self):
        return self.cartesianBounds["xMinimum"].value()

    @property
    def xMaximum(self):
        return self.cartesianBounds["xMaximum"].value()

    @property
    def yMinimum(self):
        return self.cartesianBounds["yMinimum"].value()

    @property
    def yMaximum(self):
        return self.cartesianBounds["yMaximum"].value()

    @property
    def zMinimum(self):
        return self.cartesianBounds["zMinimum"].value()

    @property
    def zMaximum(self):
        return self.cartesianBounds["zMaximum"].value()

    @property
    def sphericalBounds(self):
        return self["sphericalBounds"]

    @property
    def rangeMinimum(self):
        return self.sphericalBounds["rangeMinimum"].value()

    @property
    def rangeMaximum(self):
        return self.sphericalBounds["rangeMaximum"].value()

    @property
    def elevationMinimum(self):
        return self.sphericalBounds["elevationMinimum"].value()
    
    @property
    def elevationMaximum(self):
        return self.sphericalBounds["elevationMaximum"].value()
    
    @property
    def azimuthStart(self):
        return self.sphericalBounds["azimuthStart"].value()
    
    @property
    def azimuthEnd(self):
        return self.sphericalBounds["azimuthEnd"].value()
    
    @property
    def pose(self):
        return self["pose"]

    @property
    def acquisitionStart(self):
        return self["acquisitionStart"]

    @property
    def acquisitionStart_dateTimeValue(self):
        return self.acquisitionStart["dateTimeValue"].value()

    @property
    def acquisitionStart_isAtomicClockReferenced(self):
        return self.acquisitionStart["isAtomicClockReferenced"].value()

    @property
    def acquisitionEnd(self):
        return self["acquisitionEnd"]

    @property
    def acquisitionEnd_dateTimeValue(self):
        return self.acquisitionEnd["dateTimeValue"].value()

    @property
    def acquisitionEnd_isAtomicClockReferenced(self):
        return self.acquisitionEnd["isAtomicClockReferenced"].value()

    @property
    def pointGroupingSchemes(self):
        return self["pointGroupingSchemes"]

    @property
    def points(self):
        return self["points"]



import os
import platform
import shutil
import sys
from pathlib import Path

# Available at setup time due to pyproject.toml
from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup

HERE = Path(__file__).parent

about = {}
with open(HERE / "src" / "pye57" / "__version__.py") as f:
    exec(f.read(), about)
version = about["__version__"]

libE57_cpp = sorted(map(str, (HERE / "libE57Format" / "src").glob("*.cpp")))

libraries = []
library_dirs = []
include_dirs = [
    "libE57Format/include",
    "libE57Format/src",
    "libE57Format/extern/CRCpp/inc",
]
package_data = []

if platform.system() == "Windows":
    libraries.append("xerces-c_3")

    # using conda
    conda_library_dir = Path(sys.executable).parent / "Library"
    if conda_library_dir.exists():
        library_dirs.append(str(conda_library_dir / "lib"))
        include_dirs.append(str(conda_library_dir / "include"))

    # using cibuildwheel
    xerces_dir = Path(os.environ["TEMP"]) / "xerces_c"
    if xerces_dir.exists():
        library_dirs.append(str(xerces_dir / "lib"))
        include_dirs.append(str(xerces_dir / "include"))
        # include xerces-c dll in the package
        shutil.copy2(xerces_dir / "bin" / "xerces-c_3_2.dll", HERE / "src" / "pye57")
        package_data.append("xerces-c_3_2.dll")

else:
    libraries.append("xerces-c")

ext_modules = [
    Pybind11Extension(
        "pye57.libe57",
        ["src/pye57/libe57_wrapper.cpp"] + libE57_cpp,
        define_macros=[("E57_DLL", "")],
        include_dirs=include_dirs,
        libraries=libraries,
        library_dirs=library_dirs,
        language="c++",
    ),
]

export_header_path = HERE / "libE57Format" / "include" / "E57Export.h"


class BuildExt(build_ext):
    """A custom build extension for adding compiler-specific options."""

    def build_extensions(self):
        ct = self.compiler.compiler_type
        opts = []
        revision_id = "pye57-" + version
        if ct == "unix":
            opts.append(f'-DVERSION_INFO="{version}"')
            opts.append(f'-DREVISION_ID="{revision_id}"')
            opts.append("-DCRCPP_USE_CPP11")
            opts.append("-DCRCPP_BRANCHLESS")
            opts.append("-Wno-unused-variable")
        elif ct == "msvc":
            opts.append(f'/DVERSION_INFO="{version}"')
            opts.append(rf'/DREVISION_ID="\"{revision_id}\""')
            opts.append("/DCRCPP_USE_CPP11")
            opts.append("/DCRCPP_BRANCHLESS")
            opts.append("/DWINDOWS")
        for ext in self.extensions:
            ext.extra_compile_args = opts

        export_header_path.touch()
        try:
            super().build_extensions()
        finally:
            export_header_path.unlink()


with open(HERE / "README.md") as f:
    long_description = "\n" + f.read()

setup(
    name="pye57",
    version=version,
    author="David Caron",
    author_email="dcaron05@gmail.com",
    url="https://www.github.com/davidcaron/pye57",
    description="Python .e57 files reader/writer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=["numpy", "pyquaternion"],
    ext_modules=ext_modules,
    packages=["pye57"],
    package_dir={"": "src"},
    # include_package_data=True,
    package_data={"pye57": package_data},
    extras_require={"test": "pytest"},
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    cmdclass={"build_ext": BuildExt},
    zip_safe=False,
    python_requires=">=3.8",
)



import uuid
import os
from typing import Dict
from enum import Enum

import numpy as np
from pyquaternion import Quaternion

from pye57.__version__ import __version__
from pye57 import libe57
from pye57 import ScanHeader
from pye57.utils import convert_spherical_to_cartesian

try:
    from exceptions import WindowsError
except ImportError:
    class WindowsError(OSError):
        pass


SUPPORTED_CARTESIAN_POINT_FIELDS = {
    "cartesianX": "d",
    "cartesianY": "d",
    "cartesianZ": "d",
}

SUPPORTED_SPHERICAL_POINT_FIELDS = {
    "sphericalRange": "d",
    "sphericalAzimuth": "d",
    "sphericalElevation": "d",
}

class COORDINATE_SYSTEMS(Enum):
    CARTESIAN = SUPPORTED_CARTESIAN_POINT_FIELDS
    SPHERICAL = SUPPORTED_SPHERICAL_POINT_FIELDS

SUPPORTED_POINT_FIELDS = {
    **SUPPORTED_CARTESIAN_POINT_FIELDS,
    **SUPPORTED_SPHERICAL_POINT_FIELDS,
    "intensity": "f",
    "colorRed": "B",
    "colorGreen": "B",
    "colorBlue": "B",
    "rowIndex": "H",
    "columnIndex": "H",
    "cartesianInvalidState": "b",
    "sphericalInvalidState": "b",
}


class E57:
    def __init__(self, path, mode="r"):
        if mode not in "rw":
            raise ValueError("Only 'r' and 'w' modes are supported")
        self.path = path
        try:
            self.image_file = libe57.ImageFile(path, mode)
            if mode == "w":
                self.write_default_header()
        except Exception as e:
            try:
                self.image_file.close()
                os.remove(path)
            except (AttributeError, WindowsError, PermissionError):
                pass
            raise e

    def __del__(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if hasattr(self, "image_file"):
            self.image_file.close()

    @property
    def root(self):
        return self.image_file.root()

    @property
    def data3d(self):
        return self.root["data3D"]

    @property
    def scan_count(self):
        return len(self.data3d)

    def get_header(self, index):
        return ScanHeader(self.data3d[index])

    def write_default_header(self):
        imf = self.image_file
        imf.extensionsAdd("", libe57.E57_V1_0_URI)
        self.root.set("formatName", libe57.StringNode(imf, "ASTM E57 3D Imaging Data File"))
        self.root.set("guid", libe57.StringNode(imf, "{%s}" % uuid.uuid4()))
        self.root.set("versionMajor", libe57.IntegerNode(imf, libe57.E57_FORMAT_MAJOR))
        self.root.set("versionMinor", libe57.IntegerNode(imf, libe57.E57_FORMAT_MINOR))
        self.root.set("e57LibraryVersion", libe57.StringNode(imf, libe57.E57_LIBRARY_ID))
        self.root.set("coordinateMetadata", libe57.StringNode(imf, ""))
        creation_date_time = libe57.StructureNode(imf)
        creation_date_time.set("dateTimeValue", libe57.FloatNode(imf, 0.0))
        creation_date_time.set("isAtomicClockReferenced", libe57.IntegerNode(imf, 0))
        self.root.set("creationDateTime", creation_date_time)
        self.root.set("data3D", libe57.VectorNode(imf, True))
        self.root.set("images2D", libe57.VectorNode(imf, True))

    def make_buffer(self, field_name, capacity, do_conversion=True, do_scaling=True):
        if field_name not in SUPPORTED_POINT_FIELDS:
            raise ValueError("Unsupported point field: %s" % field_name)

        np_array = np.empty(capacity, SUPPORTED_POINT_FIELDS[field_name])
        buffer = libe57.SourceDestBuffer(self.image_file,
                                         field_name,
                                         np_array,
                                         capacity,
                                         do_conversion,
                                         do_scaling)
        return np_array, buffer

    def make_buffers(self, field_names, capacity, do_conversion=True, do_scaling=True):
        data = {}
        buffers = libe57.VectorSourceDestBuffer()
        for field in field_names:
            d, b = self.make_buffer(field, capacity, do_conversion=do_conversion, do_scaling=do_scaling)
            data[field] = d
            buffers.append(b)
        return data, buffers

    def read_scan_raw(self, index) -> Dict:
        header = self.get_header(index)
        data = {}
        buffers = libe57.VectorSourceDestBuffer()
        for field in header.point_fields:
            np_array, buffer = self.make_buffer(field, header.point_count)
            data[field] = np_array
            buffers.append(buffer)

        header.points.reader(buffers).read()

        return data

    def scan_position(self, index):
        pt = np.array([[0, 0, 0]])
        header = self.get_header(index)
        return self.to_global(pt, header.rotation, header.translation)

    @staticmethod
    def to_global(points, rotation, translation):
        rotation_matrix = Quaternion(rotation).rotation_matrix
        return (np.dot(rotation_matrix, points.T) + translation.reshape(3, 1)).T

    def read_scan(self,
                  index,
                  *,
                  intensity=False,
                  colors=False,
                  row_column=False,
                  transform=True,
                  ignore_missing_fields=False) -> Dict:
        header = self.get_header(index)
        n_points = header.point_count

        coordinate_system = header.get_coordinate_system(COORDINATE_SYSTEMS)
        if coordinate_system is COORDINATE_SYSTEMS.CARTESIAN:
            validState = "cartesianInvalidState"
            fields = list(SUPPORTED_CARTESIAN_POINT_FIELDS.keys())
        elif coordinate_system is COORDINATE_SYSTEMS.SPHERICAL: 
            validState = "sphericalInvalidState"
            fields = list(SUPPORTED_SPHERICAL_POINT_FIELDS.keys())
        if intensity:
            fields.append("intensity")
        if colors:
            fields.append("colorRed")
            fields.append("colorGreen")
            fields.append("colorBlue")
        if row_column:
            fields.append("rowIndex")
            fields.append("columnIndex")
        fields.append(validState)

        for field in fields[:]:
            if field not in header.point_fields:
                if ignore_missing_fields:
                    fields.remove(field)
                else:
                    raise ValueError("Requested to read a field (%s) with is absent from the e57 file. "
                                     "Consider using 'ignore_missing_fields' to skip it." % field)

        data, buffers = self.make_buffers(fields, n_points)
        header.points.reader(buffers).read()

        if validState in data:
            valid = ~data[validState].astype("?")

            for field in data:
                data[field] = data[field][valid]

            del data[validState]

        if transform:
            if coordinate_system is COORDINATE_SYSTEMS.CARTESIAN:
                xyz = np.array([data["cartesianX"], data["cartesianY"], data["cartesianZ"]]).T
            elif coordinate_system is COORDINATE_SYSTEMS.SPHERICAL:
                rae = np.array([data["sphericalRange"], data["sphericalAzimuth"], data["sphericalElevation"]]).T
                # rae to xyz
                xyz = convert_spherical_to_cartesian(rae)
            # translation to global coordinates
            if header.has_pose():
                xyz = self.to_global(xyz, header.rotation, header.translation)
            data["cartesianX"] = xyz[:, 0]
            data["cartesianY"] = xyz[:, 1]
            data["cartesianZ"] = xyz[:, 2]
        return data

    def write_scan_raw(self, data: Dict, *, name=None, rotation=None, translation=None, scan_header=None):
        for field in data.keys():
            if field not in SUPPORTED_POINT_FIELDS:
                raise ValueError("Unsupported point field: %s" % field)

        if rotation is None:
            rotation = getattr(scan_header, "rotation", np.array([1, 0, 0, 0]))

        if translation is None:
            translation = getattr(scan_header, "translation", np.array([0, 0, 0]))

        if name is None:
            name = getattr(scan_header, "name", "Scan %s" % len(self.data3d))

        temperature = getattr(scan_header, "temperature", 0)
        relativeHumidity = getattr(scan_header, "relativeHumidity", 0)
        atmosphericPressure = getattr(scan_header, "atmosphericPressure", 0)

        scan_node = libe57.StructureNode(self.image_file)
        scan_node.set("guid", libe57.StringNode(self.image_file, "{%s}" % uuid.uuid4()))
        scan_node.set("name", libe57.StringNode(self.image_file, name))
        scan_node.set("temperature", libe57.FloatNode(self.image_file, temperature))
        scan_node.set("relativeHumidity", libe57.FloatNode(self.image_file, relativeHumidity))
        scan_node.set("atmosphericPressure", libe57.FloatNode(self.image_file, atmosphericPressure))
        scan_node.set("description", libe57.StringNode(self.image_file, "pye57 v%s" % __version__))

        n_points = data["cartesianX"].shape[0]

        ibox = libe57.StructureNode(self.image_file)
        if "rowIndex" in data and "columnIndex" in data:
            min_row = np.min(data["rowIndex"])
            max_row = np.max(data["rowIndex"])
            min_col = np.min(data["columnIndex"])
            max_col = np.max(data["columnIndex"])
            ibox.set("rowMinimum", libe57.IntegerNode(self.image_file, min_row))
            ibox.set("rowMaximum", libe57.IntegerNode(self.image_file, max_row))
            ibox.set("columnMinimum", libe57.IntegerNode(self.image_file, min_col))
            ibox.set("columnMaximum", libe57.IntegerNode(self.image_file, max_col))
        else:
            ibox.set("rowMinimum", libe57.IntegerNode(self.image_file, 0))
            ibox.set("rowMaximum", libe57.IntegerNode(self.image_file, n_points - 1))
            ibox.set("columnMinimum", libe57.IntegerNode(self.image_file, 0))
            ibox.set("columnMaximum", libe57.IntegerNode(self.image_file, 0))
        ibox.set("returnMinimum", libe57.IntegerNode(self.image_file, 0))
        ibox.set("returnMaximum", libe57.IntegerNode(self.image_file, 0))
        scan_node.set("indexBounds", ibox)

        if "intensity" in data:
            int_min = getattr(scan_header, "intensityMinimum", np.min(data["intensity"]))
            int_max = getattr(scan_header, "intensityMaximum", np.max(data["intensity"]))
            intbox = libe57.StructureNode(self.image_file)
            intbox.set("intensityMinimum", libe57.FloatNode(self.image_file, int_min))
            intbox.set("intensityMaximum", libe57.FloatNode(self.image_file, int_max))
            scan_node.set("intensityLimits", intbox)

        color = all(c in data for c in ["colorRed", "colorGreen", "colorBlue"])
        if color:
            colorbox = libe57.StructureNode(self.image_file)
            colorbox.set("colorRedMinimum", libe57.IntegerNode(self.image_file, 0))
            colorbox.set("colorRedMaximum", libe57.IntegerNode(self.image_file, 255))
            colorbox.set("colorGreenMinimum", libe57.IntegerNode(self.image_file, 0))
            colorbox.set("colorGreenMaximum", libe57.IntegerNode(self.image_file, 255))
            colorbox.set("colorBlueMinimum", libe57.IntegerNode(self.image_file, 0))
            colorbox.set("colorBlueMaximum", libe57.IntegerNode(self.image_file, 255))
            scan_node.set("colorLimits", colorbox)

        bbox_node = libe57.StructureNode(self.image_file)
        x, y, z = data["cartesianX"], data["cartesianY"], data["cartesianZ"]
        valid = None
        if "cartesianInvalidState" in data:
            valid = ~data["cartesianInvalidState"].astype("?")
            x, y, z = x[valid], y[valid], z[valid]
        bb_min = np.array([x.min(), y.min(), z.min()])
        bb_max = np.array([x.max(), y.max(), z.max()])
        del valid, x, y, z

        if scan_header is not None:
            bb_min_scaled = np.array([scan_header.xMinimum, scan_header.yMinimum, scan_header.zMinimum])
            bb_max_scaled = np.array([scan_header.xMaximum, scan_header.yMaximum, scan_header.zMaximum])
        else:
            bb_min_scaled = self.to_global(bb_min.reshape(-1, 3), rotation, translation)[0]
            bb_max_scaled = self.to_global(bb_max.reshape(-1, 3), rotation, translation)[0]

        bbox_node.set("xMinimum", libe57.FloatNode(self.image_file, bb_min_scaled[0]))
        bbox_node.set("xMaximum", libe57.FloatNode(self.image_file, bb_max_scaled[0]))
        bbox_node.set("yMinimum", libe57.FloatNode(self.image_file, bb_min_scaled[1]))
        bbox_node.set("yMaximum", libe57.FloatNode(self.image_file, bb_max_scaled[1]))
        bbox_node.set("zMinimum", libe57.FloatNode(self.image_file, bb_min_scaled[2]))
        bbox_node.set("zMaximum", libe57.FloatNode(self.image_file, bb_max_scaled[2]))
        scan_node.set("cartesianBounds", bbox_node)

        if rotation is not None and translation is not None:
            pose_node = libe57.StructureNode(self.image_file)
            scan_node.set("pose", pose_node)
            rotation_node = libe57.StructureNode(self.image_file)
            rotation_node.set("w", libe57.FloatNode(self.image_file, rotation[0]))
            rotation_node.set("x", libe57.FloatNode(self.image_file, rotation[1]))
            rotation_node.set("y", libe57.FloatNode(self.image_file, rotation[2]))
            rotation_node.set("z", libe57.FloatNode(self.image_file, rotation[3]))
            pose_node.set("rotation", rotation_node)
            translation_node = libe57.StructureNode(self.image_file)
            translation_node.set("x", libe57.FloatNode(self.image_file, translation[0]))
            translation_node.set("y", libe57.FloatNode(self.image_file, translation[1]))
            translation_node.set("z", libe57.FloatNode(self.image_file, translation[2]))
            pose_node.set("translation", translation_node)

        start_datetime = getattr(scan_header, "acquisitionStart_dateTimeValue", 0)
        start_atomic = getattr(scan_header, "acquisitionStart_isAtomicClockReferenced", False)
        end_datetime = getattr(scan_header, "acquisitionEnd_dateTimeValue", 0)
        end_atomic = getattr(scan_header, "acquisitionEnd_isAtomicClockReferenced", False)
        acquisition_start = libe57.StructureNode(self.image_file)
        scan_node.set("acquisitionStart", acquisition_start)
        acquisition_start.set("dateTimeValue", libe57.FloatNode(self.image_file, start_datetime))
        acquisition_start.set("isAtomicClockReferenced", libe57.IntegerNode(self.image_file, start_atomic))
        acquisition_end = libe57.StructureNode(self.image_file)
        scan_node.set("acquisitionEnd", acquisition_end)
        acquisition_end.set("dateTimeValue", libe57.FloatNode(self.image_file, end_datetime))
        acquisition_end.set("isAtomicClockReferenced", libe57.IntegerNode(self.image_file, end_atomic))

        # todo: pointGroupingSchemes

        points_prototype = libe57.StructureNode(self.image_file)

        is_scaled = False
        precision = libe57.E57_DOUBLE if is_scaled else libe57.E57_SINGLE

        center = (bb_max + bb_min) / 2

        chunk_size = 5000000

        x_node = libe57.FloatNode(self.image_file, center[0], precision, bb_min[0], bb_max[0])
        y_node = libe57.FloatNode(self.image_file, center[1], precision, bb_min[1], bb_max[1])
        z_node = libe57.FloatNode(self.image_file, center[2], precision, bb_min[2], bb_max[2])
        points_prototype.set("cartesianX", x_node)
        points_prototype.set("cartesianY", y_node)
        points_prototype.set("cartesianZ", z_node)

        field_names = ["cartesianX", "cartesianY", "cartesianZ"]

        if "intensity" in data:
            intensity_min = np.min(data["intensity"])
            intensity_max = np.max(data["intensity"])
            intensity_node = libe57.FloatNode(self.image_file, intensity_min, precision, intensity_min, intensity_max)
            points_prototype.set("intensity", intensity_node)
            field_names.append("intensity")

        if all(color in data for color in ["colorRed", "colorGreen", "colorBlue"]):
            points_prototype.set("colorRed", libe57.IntegerNode(self.image_file, 0, 0, 255))
            points_prototype.set("colorGreen", libe57.IntegerNode(self.image_file, 0, 0, 255))
            points_prototype.set("colorBlue", libe57.IntegerNode(self.image_file, 0, 0, 255))
            field_names.append("colorRed")
            field_names.append("colorGreen")
            field_names.append("colorBlue")

        if "rowIndex" in data and "columnIndex" in data:
            min_row = np.min(data["rowIndex"])
            max_row = np.max(data["rowIndex"])
            min_col = np.min(data["columnIndex"])
            max_col = np.max(data["columnIndex"])
            points_prototype.set("rowIndex", libe57.IntegerNode(self.image_file, 0, min_row, max_row))
            field_names.append("rowIndex")
            points_prototype.set("columnIndex", libe57.IntegerNode(self.image_file, 0, min_col, max_col))
            field_names.append("columnIndex")

        if "cartesianInvalidState" in data:
            min_state = np.min(data["cartesianInvalidState"])
            max_state = np.max(data["cartesianInvalidState"])
            points_prototype.set("cartesianInvalidState", libe57.IntegerNode(self.image_file, 0, min_state, max_state))
            field_names.append("cartesianInvalidState")

        # other fields
        # // "sphericalRange"
        # // "sphericalAzimuth"
        # // "sphericalElevation"
        # // "timeStamp"
        # // "sphericalInvalidState"
        # // "isColorInvalid"
        # // "isIntensityInvalid"
        # // "isTimeStampInvalid"

        arrays, buffers = self.make_buffers(field_names, chunk_size)

        codecs = libe57.VectorNode(self.image_file, True)
        points = libe57.CompressedVectorNode(self.image_file, points_prototype, codecs)
        scan_node.set("points", points)

        self.data3d.append(scan_node)

        writer = points.writer(buffers)

        current_index = 0
        while current_index != n_points:
            current_chunk = min(n_points - current_index, chunk_size)

            for type_ in SUPPORTED_POINT_FIELDS:
                if type_ in arrays:
                    arrays[type_][:current_chunk] = data[type_][current_index:current_index + current_chunk]

            writer.write(current_chunk)

            current_index += current_chunk

        writer.close()
